#include <stdio.h>
#include <math.h>
#include <stdlib.h>

/**************************************************************************************************
	COPYRIGHT: SHIGERU SUZUKI 2012
	Commercial use of this program its revision is strictly prohibited.
	There is no warranty or guaranty of this program and its results.
	I would appreciate user's report of any bugs or errors of this program.
**************************************************************************************************/
	const double pi=3.141592;
	const double p0=101300.0;			/* absolute atmospheric pressure */
	int choice;					/* input signal selcctor choice option :
							   0: sinsoidal, 1: sweep, other: random input */
	int vvso;					/* velocity vector sum option, 0:all, 1:surface only */

	int i,j,k,n,ND,n_cycles,division,n_rows;	/* counter, counter, counter, number of chamber,
							   number of calculation cycles,
							   number of divisons in a term, c_cycles*division */
	int req_cycles;					/* required cycles for DFT analysis */

	double Q[25][25],KH[25][25],K[25][25];		/* for recursive formula, standard stiffness matrix,
							   actual stiffness matrix */
	double L[25],M[25],A[25],R[25],V[25],force[25];	/* duct length, duct mass, area, area ratio,
							   chamber volume, input force */
	double C[25];					/* damping coefficients : rev.07 */
	double MP[25], MM[25], MK[25][25];		/* modified mass with damping effect */
	double xp[25],xc[25],xn[25],sumx;		/* displacement vectors */
	double vc[25],sumv;				/* velocity vectors */
	double KS[13];					/* standard stiffness */
	double m0,radius,f0,F,phi,b,T,tempo,forte;	/* membranee mass,radius of membrane,f0,
							   input frequency, sweep speed [Hz/s],
							   amplitude of input, time, force */
	double Ku, HCR, delta;				/* stiffness of sp unit, thermal condition,
							   time resolution */

	FILE *input;					/* simulation condition */
	FILE *displacement, *displacement_n;		/* raw displacement output & normalized displacement output */
	FILE *dft_v;					/* sum of normalized velocity without time column for DFT */
	FILE *velocity, *velocity_n;			/* velocity of each mass & normalized velocity of each mass */
	FILE *matrix;					/* matrices K, RKR, M-1(RKR), Q */

/****** definition of normalizing: displacement & velocity of membrane are set reference value then
	ratio of area of each duct is multiplied to each displacement & velocity values 		******/

	int results(void),matrices(void);		/* avoid compile error */

int results(void)
{
//	THIS FUNCTION SEQUENTIALLY CALCULATE AND WRITE VECTORS

//	WRITE HEADER ROW

//	HEADER FOR RAW VECTOR OUTPUT

	fprintf(displacement,"time[s],force[N]");
	fprintf(velocity,"time[s],force[N]");
	for (i=0;i<ND;i++){
		fprintf(displacement,",x[%d]",i);
		fprintf(velocity,",v[%d]",i);
	}
	if(choice==0){
		fprintf(displacement,",%f[Hz]\n",F);
		fprintf(velocity,",%f[Hz]\n",F);
	}
	else
	{
	fprintf(displacement,",random\n");
	fprintf(velocity,",random\n");
	}	

//	HEADER FOR NORMALIZED VECTOR OUTPUT

	fprintf(displacement_n,"time[s],sum_x[mm]");
	fprintf(velocity_n,"time[s],sum_v[m/s]");

	for (i=0;i<ND;i++){
		fprintf(displacement_n,",x_n[%d]",i);
		fprintf(velocity_n,",v_n[%d]",i);
	}
	if(choice==0){						/* sinusoidal wave */
		fprintf(displacement_n,",force[N],%f[Hz]\n",F);
		fprintf(velocity_n,",force[N],%f[Hz]\n",F);
	}
	else if(choice==1){					/* sweep signal */
		fprintf(displacement_n,",force[N],sweep rate[Hz/s]=%f,final frequency[Hz]=%f\n",phi,F);
		fprintf(velocity_n,",force[N],sweep rate[Hz/s],final frequency[Hz]\n",phi,F);
	}
	else{							/* random signal */
		fprintf(displacement_n,",force[N],random\n");
		fprintf(velocity_n,",force[N],random\n");
	}

//	WRITE NUMBER OF POINTS AND DELTA T FOR DFT ANALYSIS

	fprintf(dft_v,"%d\n%f\n",n_rows,delta);	/* 1st line: data points, 2nd line: delta t */

//	END OF HEADER PROCESSES

	printf("Header rows were written!\n");			/* notice only */
	printf("Now recursive calculation process begins.\n");

//	RECURSIVE STEP CALCULATION BEGINS

	for (k=0;k<=n_rows;k++){

	tempo=delta*k;						/* tempo is "time" */

//	SOURCE SIGNAL OPTION (0:SINUSOIDAL, 1:LINEAR SWEEP, 2:RANDOM)

	if(choice==0){
		forte=b*sin(2*pi*F*tempo);			/* sinusoidal signal input */
	}
	else if(choice==1){
		forte=b*sin(2*pi*phi*tempo*tempo);		/* sweep signal input */
	}
	else
	{
		forte=b*2*(rand()%100-50)/100;			/* random signal input */
	}

	force[0]=2*delta*delta/MP[0]*forte;			/* force is converted to recursive formula */
	

//	VECTOR SUM CALCULATION

	sumx=0;							/* initial value */
	sumv=0;

	for (j=0;j<ND;j++){sumx=sumx+xc[j]*R[j]*1E3;}		/* sum of normalized displacement */

//
		for (i=0;i<ND;i++){
			xn[i]=0.0;
			for (j=0;j<ND;j++){
			xn[i]=xn[i]+Q[i][j]*xc[j];
			}
			xn[i]=xn[i]-MM[i]/MP[i]*xp[i]+force[i];
			vc[i]=(xc[i]-xp[i])/delta;
		}

//	CHOOSE IF INTERNAL DUCTS ARE IGNORED

	if(vvso==0){
		vvso=0;						/* all ducts are used */
	}
	else{
		vvso=n;						/* internal ducts are ignored */
	}

	sumv=sumv+vc[0]*R[0];					/* velocity of membrane */
	for (j=vvso+1;j<ND;j++){sumv=sumv+vc[j]*R[j];}		/* velocity of ducts; internal ducts are ignored if vvso=1 */


//	WRITE RAW & NORMALIZED VECTORS

	fprintf(displacement,"%f,%f",tempo,forte);
	fprintf(displacement_n,"%f,%f",tempo,sumx);
	fprintf(velocity,"%f,%f",tempo,forte);
	fprintf(velocity_n,"%f,%f",tempo,sumv);

	for (i=0;i<ND;i++){

		fprintf(displacement,",%f",xc[i]*1E3);		/* raw vector */
		fprintf(displacement_n,",%f",xc[i]*R[i]*1E3);	/* normalized vector */
		fprintf(velocity,",%f",vc[i]);			/* raw vector (m/s)*/
		fprintf(velocity_n,",%f",vc[i]*R[i]);		/* normalized vector (m/s)*/

		xp[i]=xc[i];
		xc[i]=xn[i];

	}

	fprintf(displacement,"\n");
	fprintf(displacement_n,",%f\n",forte);			/* force vector is placed at last column */

	fprintf(velocity,"\n");
	fprintf(velocity_n,",%f\n",forte);			/* force vector is placed at last column */
	fprintf(dft_v,"%f\n",sumv);				/* output for dft analysis */
	}

	return(1);
}

int matrices(void)
{
//	WRITE MATRIX VALUES FOR DEBUGGING PURPOSE ONLY

//	print "i" as header line

	for (i=0;i<ND;i++){
		fprintf(matrix,"%d,",i);
	}
	fprintf(matrix,"\n\n");

//	STANDARD STIFFNESS MATRIX

	fprintf(matrix,"standard stiffness matrix\n");
	for (i=0;i<ND;i++){
		for (j=0;j<ND;j++){
			fprintf(matrix,"%f,",KH[i][j]);
		}
	fprintf(matrix,"\n");
	}
	fprintf(matrix,"\n");

//	ACTUAL STIFFNESS MATRIX

	fprintf(matrix,"actual stiffness matrix\n");

	for (i=0;i<ND;i++){
		for (j=0;j<ND;j++){
			fprintf(matrix,"%f,",K[i][j]);
		}
	fprintf(matrix,"\n");
	}
	fprintf(matrix,"\n");

//	Q MATRIX - INTERMEDIATE MATRIX FOR RECURSIVE CALCULATION -

	fprintf(matrix,"Q matrix\n");
	for (i=0;i<ND;i++){
		for (j=0;j<ND;j++){
			fprintf(matrix,"%f,",Q[i][j]);
		}
	fprintf(matrix,"\n");
	}
	return(1);
}
main()
{
	input=fopen("input_parameters.txt","r");		/* input parameter file */

	displacement=fopen("x_vectors.csv","w");		/* raw displacement vector */
	displacement_n=fopen("normalized_x_vectors.csv","w");	/* normalized displacement vector */

	velocity=fopen("v_vectors.csv","w");			/* raw velocity vector */
	velocity_n=fopen("normalized_v_vectors.csv","w");	/* normalized velocity vector */
	dft_v=fopen("dftv.txt","w");				/* velocity output for DFT */
	
	matrix=fopen("matrices.csv","w");			/* stiffness and other matrices */

//	READ INPUT PARAMETERS FROM FILE

	fscanf(input,"%lf\t%lf\t%lf\n",&m0,&radius,&f0);	/* unit's spec */
	fscanf(input,"%d\t%d\t%d\t%lf\t%lf",&n,&n_cycles,&division,&HCR,&b);

//	PRIMARY VALUES CALCULATION IN OTDER TO READ OTHER VALUES FROM INPUT CONDITION FILE

	ND=2*n+1;						/* number of degrees of freedom */

//	READ REMAINING INPUT PARAMETERS FROM FILE

	for (i=0;i<=n;i++){fscanf(input,"%lf\t",&V[i]);}	/* volume of chambers */
	for (i=1;i<ND;i++){fscanf(input,"%lf\t",&A[i]);}	/* area of ducts */
	for (i=1;i<ND;i++){fscanf(input,"%lf\t",&L[i]);}	/* length of ducts */
	for (i=0;i<ND;i++){fscanf(input,"%lf\t",&C[i]);}	/* damping coefficients */
								/* n, cycles, heat capacity ratio, force */

//	INTERMEDIATE PARAMETER VALUES

	A[0]=pi*pow(radius,2);					/* area of membrane */
	M[0]=m0*1.0E-3;						/* mass of membrane unit conversion to [kg] */
	L[0]=m0*1.0E-3/A[0]*10000.0/1.2*1000;			/* not used anywhere but put into memory */
	R[0]=1.0;						/* A[0]/A[0]=1 */
	Ku=4.0*pow(pi,2)*pow(f0,2)*M[0];			/* spring constant of sp unit */

printf("[0] values and parameters caluculation done!\n");

//	CHOOSE USING ALL DUCT OR IGNORING INTERNAL DUCTS FOR VELOCITY VECTOR SUM CALCULATION

	printf("Choose vector sum option for later Fourier Transform analysis.\n");
	printf("\t 0: All ducts, 1: ignore internal ducts\n");
	printf("Your choice =\t");
	scanf("%d",&vvso);


//	CHOOSE INPUT SIGNAL OPTIONS

	printf("Choose Sinusoidal, Sweep or Random\nSweep option may take long to calculate.\n");
	printf("\t0:Sinusoidal, 1:Linear Sweep, 2:Random\n");
	printf("Your choice =\t");
	scanf("%d",&choice);
	printf("\n");

	if(choice==0){
		printf("input frequency[Hz]=");
		scanf("%lf",&F);			/* external force frequency */
		req_cycles=F;				/* compute required cycles */
		n_rows=n_cycles*division;		/* number of data rows */
		T=1/F;					/* term of wave in [s]*/
		delta=T/division;			/* time resolution */

	printf("You have chosen sinsoidal wave at %f [Hz].\n",F);
	}
	else if(choice==1){				/* processing sweep signal */
		printf("Highest sweep frequency is limited to 1000Hz.\n");
		printf("Maximum frequency[Hz]=");
		scanf("%lf",&F);

		printf("Sweep frequency range is 0[Hz] to %f[Hz]\n",F);
		printf("You are adviced to set %f as sweep rate for DFT analysis.\n",F);
		printf("Sweep rate[Hz/s] (recommended: %f)=",F);
		scanf("%lf",&phi);

		if(phi<10){phi=10;}			/* 1[Hz/s] <= sweep speed <=200[Hz/s] */
		if(phi>200){phi=200;}

		T=1/F;					/* term of wave in [s]*/
		delta=T/division;			/* time resolution */

		if(F>1000){F=1000;}			/* limit maxmum frequency is 1000[Hz] */
		if(F<100){F=100;}				/* at least sweep to 100[Hz] */

		n_rows= (int)(F/phi/delta);		/* update number of cycles to cover max frequency */

		printf("Number of Rows = %d\n",n_rows);
		printf("You have chosen linear sweep at %f [Hz/s] up to %f [Hz].\n",phi,F);
	}
	else						/* wrong values are assumed random signal was chosen */
	{
		F=100;					/* 100Hz is assumed for random signal analysis
							time resolution = 1/100/division for random */
		req_cycles=F;				/* assuming 100Hz signal */
		n_cycles=req_cycles;
		n_rows=n_cycles*division;		/* number of data rows */
		T=1/F;					/* term of wave in [s]*/
		delta=T/division;			/* time resolution */

		printf("You have chosen random signal.\n");
	}

//	}


//	chamber volume unit conversion litre to m^3

	for (i=0;i<=n;i++){V[i]=V[i]/1000.0;}

printf ("Unit conversion of chamber volume is complete!\n");

	for (i=1;i<ND;i++){
	R[i]=A[i]/A[0];						/* ratio of areas without unit conversion */
	M[i]=A[i]*1.0E-4*L[i]*1.0E-3*1.2;			/* mass vector calculation [kg] */
	}

	for (i=0;i<ND;i++){
	MP[i]=2*M[i]+delta*C[i];			/* modified mass matrix for damping compensation */
	MM[i]=2*M[i]-delta*C[i];			/* modified mass matrix for damping compensation */
	}

// check only
	printf("i\tR[i]\tA[i]\tL[i]\tM[i]\n");
	for (i=0;i<ND;i++){printf ("%d\t%f\t%f\t%f\t%f\n",i,R[i],A[i],L[i],M[i]);}
	printf("\n-- damping effects below --\n");
	for (i=0;i<ND;i++){printf ("%f\t",(1-MM[i]/MP[i]));}
	printf("\n");

printf("Ratio and mass vector calculation is complete!\n");

	for (i=0;i<=n;i++){KS[i]= HCR*A[0]*A[0]*1E-8*p0/V[i];
	printf("Standard stiffness matrix values %d\t%f\n",i,KS[i]);}
	printf("\n");

//	initial value setting to avoid bad values
	for (i=0;i<25;i++){
		xp[i]=0.0;		/* previous displacement vector */
		xc[i]=0.0;		/* current displacement vector */
		xn[i]=0.0;		/* next displacement vector */
		force[i]=0.0;		/* force vector */
//
		for (j=0;j<25;j++){
		KH[i][j]=0.0;		/* standard stiffness matrix */
		K[i][j]=0.0;		/* actual stiffness matrix */
		Q[i][j]=0.0;		/* intermediate matrix for calculation */
		}
	}
//	begin setting standard stiffness matrix

	for (i=0;i<=n;i++){
		for (j=0;j<=n;j++){
		KH[i][j]=KS[0];		/* upper left block */
		}
	}
printf("Upper left block of standard stiffness matrix values are calculated!\n");

	KH[0][0]=KH[0][0]+Ku;
	for (i=1;i<=n;i++){
		KH[i][i]=KH[i][i]+KS[i];	/* diagonal element of upper left block */
	}
	for (i=1;i<=n;i++){
		KH[n+i][i]=-KS[i];	/* lower left block */
		KH[i][n+i]=-KS[i];	/* upper right block */
	}
	for (i=1;i<=n;i++){
		KH[n+i][n+i]=KS[i];	/* lower right block */
	}
//	end setting standard stiffness matrix

printf("Standard stiffnes matrix became ready!\n");

//	begin calculating actual stiffness matrix
	for (i=0;i<ND;i++){
		for (j=0;j<ND;j++){
			K[i][j]=KH[i][j]*R[i]*R[j];
		}
	}
//	stiffness matrix complete
printf("Actual stiffnes matrix became ready!\n\n");


//	MK[i][j]=2*(2*M[i]-delta^2*K[i][j]) rev07 and later
//	 -- this matrix is to calculate coefficient matrix of x[j] --

	for(i=0;i<ND;i++){for(j=0;j<ND;j++){MK[i][j]=0.0;}}	/* initial value = 0 */

	for(i=0;i<ND;i++){
		MK[i][i]=M[i];					/* diagonal elements are given */
	}
	for(i=0;i<ND;i++){
		for(j=0;j<ND;j++){
			MK[i][j]=2*(2*MK[i][j]-pow(delta,2)*K[i][j]); /* all elements */
		}
	}

//	Calculate "Q=MK/MP" rev07 and later
//	 -- this is coefficient matrix of x[j] --

	for (i=0;i<ND;i++){
		for (j=0;j<ND;j++){
		Q[i][j]=MK[i][j]/MP[i];
		}
	}

printf("Ready to start recursive calculation!\n\n");

	results();	/* calculate recursive formulae and save in a file */

printf("Dynamic simulation has been complete!\n\n");

	matrices();	/* save matrix files for debugging */
printf("Matrces have been recorded.\n\n");

fclose(input);
fclose(displacement);
fclose(displacement_n);
fclose(velocity);
fclose(velocity_n);
fclose(matrix);
fclose(dft_v);
}
